/*****************************************************************************/
/*!
 * \file expr_value.cpp
 * 
 * Author: Sergey Berezin
 * 
 * Created: Fri Feb  7 22:29:04 2003
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/

#include "expr_value.h"
#include "expr_manager.h"
#include "notifylist.h"
#include "debug.h"

using namespace std;

namespace CVCL {

////////////////////////////////////////////////////////////////////////
// Class ExprValue static members
////////////////////////////////////////////////////////////////////////

std::hash<char*> ExprValue::s_charHash;
std::hash<long int> ExprValue::s_intHash;

////////////////////////////////////////////////////////////////////////
// Class ExprValue methods
////////////////////////////////////////////////////////////////////////

// Return a ref to an empty vector (stored in ExprManager)
const std::vector<Expr>& ExprValue::getKids() const {
  return d_em->getEmptyVector();
}

// Returns the optional operator in ExprNode; it is undefined if there
// is no operator.
// const Expr& ExprValue::getOpRef() const {
//   DebugAssert(false, "ExprValue::getOp() on Expr without an op");
//   return d_em->getNullExpr();
// }

// Returns (ID "name") for uninterpreted functions and constants,
// and (BOUND_ID "name" "uniqueID") for bound variables
const string& ExprValue::getName() const {
  static string ret = "";
  DebugAssert(false, "getName() is called on ExprValue");
  return ret;
}

// Generic Values (parts of expression)
  
const Rational& ExprValue::getRatValue(int idx) const {
  static const Rational null(0);
  DebugAssert(false, "ExprValue::getRatValue("+int2string(idx)
	      +"): called on the base class");
  return null;
}


int ExprValue::getIntValue(int idx) const {
  DebugAssert(false, "ExprValue::getIntValue("+int2string(idx)
	       +"): called on the base class");
  return 0;
}

const string& ExprValue::getStringValue(int idx) const {
  DebugAssert(false, "ExprValue::getStringValue("+int2string(idx)
	       +"): called on the base class");
  static string null("");
  return null;
}

const Expr& ExprValue::getExprValue(int idx) const {
  static Expr null;
  DebugAssert(false, "ExprValue::getExprValue("+int2string(idx)
	       +"): called on the base class");
  return null;
}

const Theorem& ExprValue::getTheoremValue(int idx) const {
  static Theorem null;
  DebugAssert(false, "ExprValue::getTheoremValue("+int2string(idx)
	       +"): called on the base class");
  return null;
}

const CDO<int>& ExprValue::getCDIntValue(int idx) const {
  static CDO<int>* null = NULL;
  DebugAssert(false, "ExprValue::getCDIntValue("+int2string(idx)
	       +"): called on the base class");
  return *null;
}

const CDO<Expr>& ExprValue::getCDExprValue(int idx) const {
  static CDO<Expr>* null = NULL;
  DebugAssert(false, "ExprValue::getCDExprValue("+int2string(idx)
	       +"): called on the base class");
  return *null;
}

const CDO<Theorem>& ExprValue::getCDTheoremValue(int idx) const {
  static CDO<Theorem>* null = NULL;
  DebugAssert(false, "ExprValue::getCDTheoremValue("+int2string(idx)
	       +"): called on the base class");
  return *null;
}

const void* ExprValue::getOtherValue(int idx) const {
  DebugAssert(false, "ExprValue::getOtherValue("+int2string(idx)
	       +"): called on the base class");
  return (void*)0;
}

  // Generic attributes
int& ExprValue::getIntAttr(int idx) {
  static int null(0);
  DebugAssert(false, "ExprValue::getIntAttr("+int2string(idx)
	       +"): called on the base class");
  return null;
}

Expr& ExprValue::getExprAttr(int idx) {
  static Expr null;
  DebugAssert(false, "ExprValue::getExprAttr("+int2string(idx)
	       +"): called on the base class");
  return null;
}

Theorem& ExprValue::getTheoremAttr(int idx) {
  static Theorem null;
  DebugAssert(false, "ExprValue::getTheoremAttr("+int2string(idx)
	       +"): called on the base class");
  return null;
}

CDO<int>& ExprValue::getCDIntAttr(int idx) {
  static CDO<int>* null = NULL;
  DebugAssert(false, "ExprValue::getCDIntAttr("+int2string(idx)
	       +"): called on the base class");
  return *null;
}

CDO<Expr>& ExprValue::getCDExprAttr(int idx) {
  static CDO<Expr>* null = NULL;
  DebugAssert(false, "ExprValue::getCDExprAttr("+int2string(idx)
	       +"): called on the base class");
  return *null;
}

CDO<Theorem>& ExprValue::getCDTheoremAttr(int idx) {
  static CDO<Theorem>* null = NULL;
  DebugAssert(false, "ExprValue::getCDTheoremAttr("+int2string(idx)
	       +"): called on the base class");
  return *null;
}

void* ExprValue::getOtherAttr(int idx) {
  DebugAssert(false, "ExprValue::getOtherAttr("+int2string(idx)
	       +"): called on the base class");
  return (void*)0;
}


// Descructor
ExprValue::~ExprValue() {
  // Be careful deleting the attributes: first set them to NULL, then
  // delete, to avoid circular destructor calls
  if (d_find) {
    CDO<Theorem>* find = d_find;
    d_find = NULL;
    delete find;
  }
  if(d_notifyList != NULL) {
    NotifyList* nl = d_notifyList;
    d_notifyList = NULL;
    delete nl;
  }
  // Set all smart pointers to Null
  d_type = Type();
  d_tcc = Expr();
  d_subtypePred=Theorem();
  d_simpCache=Theorem();
  d_simpFrom=Expr();
}

// Equality between any two ExprValue objects (including subclasses)
bool ExprValue::operator==(const ExprValue& ev2) const {
  DebugAssert(getMMIndex() == EXPR_VALUE,
	      "ExprValue::operator==() called from a subclass");
  if(getMMIndex() != ev2.getMMIndex())
    return false;

  return (getKind() == ev2.getKind());
}


ExprValue* ExprValue::copy(ExprManager* em, ExprIndex idx) const {
  DebugAssert(getMMIndex() == EXPR_VALUE,
	      "ExprValue::copy() called from a subclass");
  return new(em->getMM(EXPR_VALUE)) ExprValue(em, d_kind, idx);
}


ExprValue* ExprValue::copy(ExprManager* em, const vector<Expr>& kids,
			   ExprIndex idx) const {
  DebugAssert(kids.size()==0, "ExprValue::copy(kids)");
  return copy(em, idx);
}


bool ExprNode::operator==(const ExprValue& ev2) const {
  if(getMMIndex() != ev2.getMMIndex())
    return false;

  return (getKind() == ev2.getKind())
    && (getKids() == ev2.getKids());
}

ExprValue* ExprNode::copy(ExprManager* em, ExprIndex idx) const {
  return new(em->getMM(getMMIndex())) ExprNode(em, d_kind, d_children, idx);
}


ExprValue* ExprNode::copy(ExprManager* em, const std::vector<Expr>& kids,
			  ExprIndex idx) const {
  return new(em->getMM(getMMIndex())) ExprNode(em, d_kind, kids, idx);
}


bool ExprString::operator==(const ExprValue& ev2) const {
  if(getMMIndex() != ev2.getMMIndex())
    return false;

  return (getString() == ev2.getString());
}

ExprValue* ExprString::copy(ExprManager* em, ExprIndex idx) const {
  return new(em->getMM(getMMIndex())) ExprString(em, d_str, idx);
}



bool ExprSkolem::operator==(const ExprValue& ev2) const {
  if(getMMIndex() != ev2.getMMIndex())
    return false;

  return (getBoundIndex() == ev2.getBoundIndex()
	  && getExistential() == ev2.getExistential());
}

ExprValue* ExprSkolem::copy(ExprManager* em, ExprIndex idx) const {
  return new(em->getMM(getMMIndex()))
    ExprSkolem(em, getBoundIndex(), getExistential(), idx);
}


ExprValue* ExprSkolem::rebuild(ExprManager* em) const {
  return new(em->getMM(getMMIndex()))
    ExprSkolem(em, getBoundIndex(), rebuildExpr(em, getExistential()));
}


bool ExprRational::operator==(const ExprValue& ev2) const {
  if(getMMIndex() != ev2.getMMIndex())
    return false;

  return (getRational() == ev2.getRational());
}

ExprValue* ExprRational::copy(ExprManager* em, ExprIndex idx) const {
  return new(em->getMM(getMMIndex())) ExprRational(em, d_r, idx);
}


bool ExprVar::operator==(const ExprValue& ev2) const {
  if(getMMIndex() != ev2.getMMIndex())
    return false;

  return (getKind() == ev2.getKind() && getName() == ev2.getName());
}

ExprValue* ExprVar::copy(ExprManager* em, ExprIndex idx) const {
  return new(em->getMM(getMMIndex())) ExprVar(em, d_name, idx);
}


bool ExprBoundVar::operator==(const ExprValue& ev2) const {
  if(getMMIndex() != ev2.getMMIndex())
    return false;

  return (getKind() == ev2.getKind() && getName() == ev2.getName()
	  && getUid() == ev2.getUid());
}

ExprValue* ExprBoundVar::copy(ExprManager* em, ExprIndex idx) const {
  return new(em->getMM(getMMIndex())) ExprBoundVar(em, d_name, d_uid, idx);
}


bool ExprCC::operator==(const ExprValue& ev2) const {
  if(getMMIndex() != ev2.getMMIndex())
    return false;

  return (getKind() == ev2.getKind())
      && (getKids() == ev2.getKids());
}

ExprValue* ExprCC::copy(ExprManager* em, ExprIndex idx) const {
  return new(em->getMM(getMMIndex())) ExprCC(em, d_kind, d_children, idx);
}


ExprValue* ExprCC::copy(ExprManager* em, const std::vector<Expr>& kids,
			ExprIndex idx) const {
  return new(em->getMM(getMMIndex())) ExprCC(em, d_kind, kids, idx);
}


bool ExprApply::operator==(const ExprValue& ev2) const {
  if(getMMIndex() != ev2.getMMIndex())
    return false;

  return (getFun() == ev2.getFun())
      && (getKids() == ev2.getKids());
}

ExprValue* ExprApply::copy(ExprManager* em, ExprIndex idx) const {
  return new(em->getMM(getMMIndex())) ExprApply(em, d_fun, d_children, idx);
}


ExprValue* ExprApply::copy(ExprManager* em, const std::vector<Expr>& kids,
			   ExprIndex idx) const {
  return new(em->getMM(getMMIndex())) ExprApply(em, d_fun, kids, idx);
}


ExprValue* ExprApply::rebuild(ExprManager* em) const {
  vector<Expr> newKids;
  rebuildExprs(em, d_children, newKids);
  return new(em->getMM(getMMIndex()))
    ExprApply(em, rebuildExpr(em, d_fun), newKids);
}


bool ExprClosure::operator==(const ExprValue& ev2) const {
  if(getMMIndex() != ev2.getMMIndex())
    return false;

  return (getKind() == ev2.getKind())
      && (getBody() == ev2.getBody())
      && (getVars() == ev2.getVars());
}

ExprValue* ExprClosure::copy(ExprManager* em, ExprIndex idx) const {
  return new(em->getMM(getMMIndex()))
    ExprClosure(em, d_kind, d_vars, d_body, idx);
}


ExprValue* ExprClosure::rebuild(ExprManager* em) const {
  vector<Expr> newVars;
  rebuildExprs(em, d_vars, newVars);
  return new(em->getMM(getMMIndex()))
    ExprClosure(em, d_kind, newVars, rebuildExpr(em, d_body));
}


////////////////////////////////////////////////////////////////////////
// Methods of subclasses of ExprValue
////////////////////////////////////////////////////////////////////////

// Hash function for subclasses with kids.
size_t ExprValue::hash(const int kind, const std::vector<Expr>& kids) {
  size_t res(s_intHash((long int)kind));
  for(std::vector<Expr>::const_iterator i=kids.begin(), iend=kids.end();
      i!=iend; ++i) {
    void* ptr = i->d_expr;
    res = res*PRIME + pointerHash(ptr);
  }
  return res;
}




// These exprs have no kids


// const vector<Expr>& ExprSkolem::getKids() const {
//   DebugAssert(false, "ExprSkolem::getKids() should not be called");
//   return d_em->getEmptyVector();
// }

// const vector<Expr>& ExprString::getKids() const {
//   DebugAssert(false, "ExprString::getKids() should not be called");
//   return d_em->getEmptyVector();
// }


const vector<Expr>& ExprRational::getKids() const {
  DebugAssert(false, "ExprRational::getKids() should not be called");
  return d_em->getEmptyVector();
}


const vector<Expr>& ExprVar::getKids() const {
  DebugAssert(false, "ExprVar::getKids() should not be called");
  return d_em->getEmptyVector();
}


const vector<Expr>& ExprBoundVar::getKids() const {
  DebugAssert(false, "ExprBoundVar::getKids() should not be called");
  return d_em->getEmptyVector();
}


ExprCC::~ExprCC()
{
  // Be careful deleting the attributes: first set them to NULL, then
  // delete, to avoid circular destructor calls
  if (d_sig) {
    CDO<Theorem>* sig = d_sig;
    d_sig = NULL;
    delete sig;
  }
  if (d_rep) {
    CDO<Theorem>* rep = d_rep;
    d_rep = NULL;
    delete rep;
  }
}


size_t ExprClosure::computeHash() const {
  return d_body.hash()*PRIME + ExprValue::hash(d_kind, d_vars);
}


// ExprRead::~ExprRead()
// {
//   // Be careful deleting the attributes: first set them to NULL, then
//   // delete, to avoid circular destructor calls
//   if (d_sig) {
//     CDO<Theorem>* sig = d_sig;
//     d_sig = NULL;
//     delete sig;
//   }
//   if (d_rep) {
//     CDO<Theorem>* rep = d_rep;
//     d_rep = NULL;
//     delete rep;
//   }
// }

// size_t
// ExprRecord::computeHash() const {
//   size_t res(ExprValue::hash(d_kind, d_children));
//   for(vector<string>::const_iterator i=d_fields.begin(), iend=d_fields.end();
//       i!=iend; ++i)
//     res = res*PRIME + s_charHash(i->c_str());
//   return res;
// }

} // end of namespace CVCL
